# Copyright (C) 2002-2021 CERN for the benefit of the ATLAS collaboration

from AthenaCommon import Logging
from ..powheg_V2 import PowhegV2

## Get handle to Athena logging
logger = Logging.logging.getLogger("PowhegControl")


class ZZ(PowhegV2):
    """! Default Powheg configuration for Z-boson pair production including interference for identical leptons.

    Create a configurable object with all applicable Powheg options.

    @author James Robinson  <james.robinson@cern.ch>
    """

    def __init__(self, base_directory, **kwargs):
        """! Constructor: all process options are set here.

        @param base_directory: path to PowhegBox code.
        @param kwargs          dictionary of arguments from Generate_tf.
        """
        super(ZZ, self).__init__(base_directory, "ZZ", **kwargs)

        # Add parameter validation functions
        self.validation_functions.append("validate_decays")

        ## List of allowed decay modes
        self.allowed_decay_modes = ["z z > e+ e- e+ e-", "z z > e+ e- tau+ tau-", "z z > j j j j", "z z > j j l+ l-",
                                    "z z > l+ l- l'+ l'-", "z z > mu+ mu- e+ e-", "z z > mu+ mu- mu+ mu-",
                                    "z z > tau+ tau- mu+ mu-", "z z > tau+ tau- tau+ tau-", "z z > v v~ e+ e-",
                                    "z z > v v~ j j", "z z > v v~ l+ l-", "z z > v v~ mu+ mu-", "z z > v v~ tau+ tau-",
                                    "z z > v v~ v' v'~"]

        # Add all keywords for this process, overriding defaults if required
        self.add_keyword("bornktmin")
        self.add_keyword("bornonly")
        self.add_keyword("bornsuppfact")
        self.add_keyword("bornzerodamp")
        self.add_keyword("bottomthr")
        self.add_keyword("bottomthrpdf")
        self.add_keyword("btildeborn")
        self.add_keyword("btildecoll")
        self.add_keyword("btildereal")
        self.add_keyword("btildevirt")
        self.add_keyword("btlscalect")
        self.add_keyword("btlscalereal")
        self.add_keyword("charmthr")
        self.add_keyword("charmthrpdf")
        self.add_keyword("check_bad_st2")
        self.add_keyword("clobberlhe")
        self.add_keyword("colltest")
        self.add_keyword("compress_lhe")
        self.add_keyword("compress_upb")
        self.add_keyword("compute_rwgt")
        self.add_keyword("cutallpairs")
        self.add_keyword("doublefsr")
        self.add_keyword("dronly")
        self.add_keyword("evenmaxrat")
        self.add_keyword("facscfact", self.default_scales[0])
        self.add_keyword("fastbtlbound")
        self.add_keyword("fixedgrid")
        self.add_keyword("fixedscale")
        self.add_keyword("flg_debug")
        self.add_keyword("foldcsi", 2)
        self.add_keyword("foldphi")
        self.add_keyword("foldy", 2)
        self.add_keyword("fullrwgt")
        self.add_keyword("fullrwgtmode")
        self.add_keyword("hdamp")
        self.add_keyword("hfact")
        self.add_keyword("icsimax")
        self.add_keyword("ih1")
        self.add_keyword("ih2")
        self.add_keyword("itmx1", 6)
        self.add_keyword("itmx1rm")
        self.add_keyword("itmx2", 6)
        self.add_keyword("itmx2rm")
        self.add_keyword("iupperfsr")
        self.add_keyword("iupperisr")
        self.add_keyword("iymax")
        self.add_keyword("lhans1", self.default_PDFs)
        self.add_keyword("lhans2", self.default_PDFs)
        self.add_keyword("lhapdf6maxsets")
        self.add_keyword("lhrwgt_descr")
        self.add_keyword("lhrwgt_group_combine")
        self.add_keyword("lhrwgt_group_name")
        self.add_keyword("lhrwgt_id")
        self.add_keyword("LOevents")
        self.add_keyword("manyseeds")
        self.add_keyword("max_io_bufsize")
        self.add_keyword("maxseeds")
        self.add_keyword("minlo")
        self.add_keyword("mintupbratlim")
        self.add_keyword("mintupbxless")
        self.add_keyword("mllmin")
        self.add_keyword("ncall1", 40000)
        self.add_keyword("ncall1rm")
        self.add_keyword("ncall2", 60000)
        self.add_keyword("ncall2rm")
        self.add_keyword("ncallfrominput")
        self.add_keyword("noevents")
        self.add_keyword("novirtual")
        self.add_keyword("nubound", 20000)
        self.add_keyword("olddij")
        self.add_keyword("par_2gsupp")
        self.add_keyword("par_diexp")
        self.add_keyword("par_dijexp")
        self.add_keyword("parallelstage")
        self.add_keyword("pdfreweight")
        self.add_keyword("ptsqmin")
        self.add_keyword("ptsupp")
        self.add_keyword("radregion")
        self.add_keyword("rand1")
        self.add_keyword("rand2")
        self.add_keyword("renscfact", self.default_scales[1])
        self.add_keyword("runningwidth")
        self.add_keyword("rwl_add")
        self.add_keyword("rwl_file")
        self.add_keyword("rwl_format_rwgt")
        self.add_keyword("rwl_group_events")
        self.add_keyword("smartsig")
        self.add_keyword("softtest")
        self.add_keyword("stage2init")
        self.add_keyword("storeinfo_rwgt")
        self.add_keyword("storemintupb")
        self.add_keyword("testplots")
        self.add_keyword("testsuda")
        self.add_keyword("ubexcess_correct")
        self.add_keyword("ubsigmadetails")
        self.add_keyword("use-old-grid")
        self.add_keyword("use-old-ubound")
        self.add_keyword("withdamp")
        self.add_keyword("withinterference")
        self.add_keyword("withnegweights")
        self.add_keyword("withsubtr")
        self.add_keyword("xgriditeration")
        self.add_keyword("xupbound", 4)
        self.add_keyword("zerowidth")
        self.add_keyword("ZZeeee", self.allowed_decay_modes[0], name="decay_mode", hidden=False)
        self.add_keyword("ZZeetautau")
        self.add_keyword("ZZllll")
        self.add_keyword("ZZmumuee")
        self.add_keyword("ZZmumumumu")
        self.add_keyword("ZZqqll")
        self.add_keyword("ZZqqqq")
        self.add_keyword("ZZtautaumumu")
        self.add_keyword("ZZtautautautau")
        self.add_keyword("ZZvvee")
        self.add_keyword("ZZvvll")
        self.add_keyword("ZZvvmumu")
        self.add_keyword("ZZvvqq")
        self.add_keyword("ZZvvtautau")
        self.add_keyword("ZZvvvv")

    def validate_decays(self):
        """! Validate idvecbos and vdecaymode keywords."""
        # See https://docs.google.com/spreadsheets/d/1Aa7FwB74ppHbXles5LyHrKGlvUFi5PxbZC-Mrc3Lz90 for meanings
        self.expose()  # convenience call to simplify syntax
        self.check_decay_mode(self.decay_mode, self.allowed_decay_modes)
        # Enable appropriate decay mode
        list(self.parameters_by_keyword("ZZeeee"))[0].value = 0  # disable the one used as a proxy for decay mode
        __decay_mode_lookup = {"z z > e+ e- e+ e-": "ZZeeee",
                               "z z > e+ e- tau+ tau-": "ZZeetautau",
                               "z z > j j j j": "ZZqqqq",
                               "z z > j j l+ l-": "ZZqqll",
                               "z z > l+ l- l'+ l'-": "ZZllll",
                               "z z > mu+ mu- e+ e-": "ZZmumuee",
                               "z z > mu+ mu- mu+ mu-": "ZZmumumumu",
                               "z z > tau+ tau- mu+ mu-": "ZZtautaumumu",
                               "z z > tau+ tau- tau+ tau-": "ZZtautautautau",
                               "z z > v v~ e+ e-": "ZZvvee",
                               "z z > v v~ j j": "ZZvvqq",
                               "z z > v v~ l+ l-": "ZZvvll",
                               "z z > v v~ mu+ mu-": "ZZvvmumu",
                               "z z > v v~ tau+ tau-": "ZZvvtautau",
                               "z z > v v~ v' v'~": "ZZvvvv"}
        list(self.parameters_by_keyword(__decay_mode_lookup[self.decay_mode]))[0].value = 1
